/*
LFMT32 -- reformat text data.
    Copyright (C) 2004  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later.  Requires comdlg32.lib odbc32.lib odbccp32.lib winmm.lib
	cl lfmt32.c /link odbc32.lib odbccp32.lib 
*/


#define MAIN
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <process.h>
#ifdef WIN32
#include <dos.h>
#include <windows.h>
#include <io.h>
#include <fcntl.h>
#endif

#define MAXLINE 65536
// #define OUTLEN  4096

#define MAXTOK 255
#define MAXSTRPARM 255
char tokbuf[MAXLINE+64];
char holdbuf[MAXLINE+64];
char holdpattern[MAXSTRPARM];

FILE *infi, *otfi;
int lineno = 0, usenewline = 1;
char inbuff[MAXLINE+MAXLINE],outbuff[MAXLINE+MAXLINE];
char cmdchar='$';
char line[MAXLINE+64];
char format[2048];
int done=0; //set to 1 if eof reached unexpectedly
extern int _fmode;

// Modifed 5/2004 to add a hold field.  If specified then any line
// in the file that matches a special pattern will be placed in a hold
// buffer.  The $H token will prefix the hold buffer to the output
// Modified 9/8/99 to fix tokenization.  Using strtok
// Modified 9/9/99 to remove strtok and use my_strtok instead
// Modified 9/13/99 to add $fn- which tells the program to take
// all of the text from field n to the end of the record and 
// write it to the output
// Modified 9/14/99 to work in binary mode so that char(26)
// are ignored.  This is now a 32 bit application.
// Modified 11/21/99 to fix the previous delim fix.  Needed special processing 
// when one of the delimiters is a space, forcing leading blanks to be 
// removed before looking for the end of field.  For all other delimiters
// this should not be done so that the program will handle empty delimited
// fields, e.g. ,,
// Modified 2000-2003 minor bug fixes dealing with delimitted fields
// Modified 1/23/2004 -- added new token, 'D' to be used to remove 
// leading and trailing whitespace from a fixed length field to make it
// easier to parse the output of database extract.
// Increased line size of the longest string from 4096 to 65536

char *trim(char *buf) {
	char *b;
	int ln,c;

	b=buf;
	// remove white space from start and end of buf
	if (buf == NULL)
		return NULL;
	// remove leading white space
	while (isspace(*b))
		++b;
	// if string is empty, return pointer to empty string
	if (*b == 0)
		return b;
	// remove trailing white space
	for (ln = strlen(b)-1; ln >= 0; --ln) {
		c = *(b+ln);
		if (isspace(c))
			*(b+ln) = 0;
		else
			return b;
	}
	return NULL;		

}

char *my_strtok(char *str, char *delim) {
	static char *starthere;
	char *s, *start;
	char *d=delim;
	char prev_delim;
	char *match_cur = NULL;
	char *match_min = NULL;

	if (str) 
		// If a new string, set starting postion.
		starthere = str;

	// start at starting position
	s = starthere;

	if (strchr(delim,' ')) {
		// One of the delimiters is a space character
		// remove whitespace at the beginning of the string
		while (*s==' ')
			++s;
	}

	if (*s == 0)
		return NULL;

	start = s;

	// Look for 1st matching delim character
	while (*d) {
		if (match_cur = strchr(s, *d++)) {
			// If *d is a space character, then look for 1st non space
			if (match_min == NULL)
				match_min = match_cur;
			if (match_min > match_cur)
				match_min = match_cur;
		}
	}

	// done scanning
	if (match_min) {
		// found a match; terminate string
		prev_delim = *match_min;
		*match_min = 0;
		++match_min;
		starthere = match_min;
		if (prev_delim == ' ' && strchr(delim,' ')) {
			// a space character is one of the delimiters
			// when this happens, look for the next non space
			// If this character is also a delim then set starthere
			// at the position past this non blank delimiter
			// otherwise use the first space as the delim
			while (*match_min == ' ')
				++match_min;
			if (*match_min && strchr(delim,*match_min))
				starthere = match_min + 1;
		}
	}
	else 
		starthere = start + strlen(start);

	return start; 
}

void get_tokens(char *buf, char *token[], char *delim) {
	int i;
	char *tok,*b;

	b = tokbuf;
	strcpy(tokbuf,buf); // Move into token buffer
	// look for a delim character, using my_strtok
	// strtok can not be used because it removes
	// multiple occurrences of delimiters concatenated
	// together
	tok = my_strtok(b,delim);
	for (i=0; tok && i < MAXTOK;++i) {
		token[i] = trim(tok);
		tok = my_strtok(NULL,delim); // other times
	}
	if (i < (MAXTOK -1))
		token[++i] = NULL;
}
		

void print_token(char *token, int offset, int len, int width, 
		 char rightjustify, FILE *otfi) {
	// Write the token to otfi
	if (offset >= len)  {
		/* don't print anything if offset past end of line */
		// If a field width was specified, fill with blanks
		while(width--)
			putc(' ',otfi);
		return;
	}
	
	if (width == 0) {
		width = len;
		if (len < (width+offset))
			width = len - offset;
		fwrite(token+offset,width,1,otfi);
		return;
	}

	// If the field length is >= w,just write the field data
	// truncate at width
	if ((len - offset) >= width)
		fwrite(token+offset,width,1,otfi);
	else {
		// write out as much data as is present, then pad
		if (rightjustify) {
		// pad first
			width -= (len - offset);
			while(width--)
				putc(' ',otfi);
		}
		// write the data
		fwrite(token+offset,len-offset,1,otfi);
		if (rightjustify == 0) {
			// pad after
			width -= (len - offset);
			while(width--)
				putc(' ',otfi);
		}
	}
	return;
}

// look for +... in string, if so, calculate the field offset
char *getfieldoffset(char *s, size_t *offset) {
	size_t o=0;
	if (*(++s) == '+') {
		++s;
	while (isdigit(*s))
		o = o * 10 + (*s++ - '0');
	}
    *offset = o;
	return s;
}


// Look for a token number following 'F', return
char *gettokennumber(char *s,int *tok) {
	*tok = 0;
	if (isdigit(*(++s))) {
		while(isdigit(*s)) 
			*tok = *tok*10 + (*s++ - '0');
		*tok --;		// Make token number relative 0, not 1
		s--;		// Make s point to last character in $f
	}
	return s;
}

void printalltokens(char *token[]) {
	int tok=0;
	// This is shorthand for f1=$f1, f2=$f2, ... ignore field width in this case
	for (tok=0;token[tok];++tok) {
		if (tok)
			fprintf(otfi,", ");
		fprintf(otfi,"$f%d=%s",tok+1,token[tok]);
	}
	fprintf(otfi,". ");
}




void expand(char *s,char *l1, char *delim) {
	int w,o,tok,last_token;
	char tokbuf[MAXLINE+1];
	int flen,len,expanding,ln;
	char *token[MAXTOK+1];
	char tokens = 0;
	char temp[7],*c;
	char *l, rightjustify=0, *start;

	l = l1; // Start using the buffer passed to expand

	if (l[(len = strlen(l))-1] == '\n')
		l[--len] = 0;
	if (l[(len = strlen(l))-1] == '\r')
		l[--len] = 0;
	ln = strlen(holdbuf);
	expanding = 0;

	// Initialize array to all empty strings
	for (tok=0; tok <= MAXTOK; token[tok++] = NULL);

	while (*s) {
		start = s;
		if (!expanding) {
			w = 0;
			o = 0;
			/* now position 1 char past % */
			c = strchr(s,cmdchar);
			if (c == NULL) { /* nothing to expand, just write the format out */
				fputs(s,otfi);
				if (usenewline)
					putc('\n',otfi);
				return;
			}
			if (c > s) { /* write out everything up to % */
				fwrite(s,1,c-s,otfi);
				s = c;
			}
			++s;
			rightjustify = 0;
		}
		expanding = 1;
		switch (toupper(*s)) {
		case 'R':
			if (fgets(line,sizeof(line),infi) == NULL) {
				done = 1;
				if (usenewline) 
					putc('\n',otfi);
				return;
			}
			else 
				l = line;
			// Remove trailing CR and LF
			if (l[(len = strlen(l))-1] == '\n')
				l[--len] = 0;
			if (l[(len = strlen(l))-1] == '\r')
				l[--len] = 0;

			++s;
			o = 0;
			tokens=0;
			expanding = 0;
			break;
		case 'Q':
			putc('\"',otfi);
			++s;
			expanding = 0;
			break;
		case 'S':
			putc('/',otfi);
			++s;
			expanding = 0;
			break;
		case 'N':
			putc('\n',otfi);
			++s;
			expanding = 0;
			break;
		case 'G':
			putc('>',otfi);
			++s;
			expanding = 0;
			break;
		case 'L':
			putc('<',otfi);
			++s;
			expanding = 0;
			break;
		case 'B':
			putc('|',otfi);
			++s;
			expanding = 0;
			break;
		case 'T':
			putc('\t',otfi);
			++s;
			expanding = 0;
			break;
		case '.':
			// If the period follows the control character ($ or %)
			// and if it precedes a digit (0..9), then set the right
			// justify flag, else print out the period.
			if (isdigit(*(s+1))) {
				rightjustify = 1;
				++s;
			}
			else {
				putc('.',otfi);
				expanding = 0;
				++s;
			}
			break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			w = 10*w + (*s - '0');
			++s;
			break;
		case ' ':
			if (w)
				while (w--)
					putc(' ',otfi);
			else
				putc(' ',otfi);
			++s;
			expanding = 0;
			break;
		// print one or more delimited fields
		// as $wFn[+o] or $wFn-m[+o}
		case 'F':
			expanding = 0;
			// find first 255 tokens if present
			// if a field width set, then ensure that the field
			// is always the specified size, even if it means
			// padding with blanks.
			if (tokens == 0) {
				get_tokens(l,token, delim);
				tokens = 1;
			}
			// This is shorthand for f1=$f1, f2=$f2, ...
			// ignore field width in this case
			s = gettokennumber(s,&tok);
			if (*s == '=') {
				printalltokens(token);
				++s;
				break;
			}
			--tok;
			if (token[tok] == NULL) {
				++s;
				while(w--)
					putc(' ',otfi);
				if (*s == '+') {
					++s;
					// skip past offset
					while (isdigit(*s))
						++s;
				}
				break; // field does not exist, write empty string
			}
			else
				flen = strlen(token[tok]);
			
			s = getfieldoffset(s,&o);

			last_token = tok;
			if (*s == '-') {
				// specified a range of fields, ie. $f1-15
				// or all the rest of the fields $f1-
				last_token = 0;
				++s;
				while (isdigit(*s))
					last_token = last_token * 10 + (*s++ - '0');
				if (last_token == 0)
					if (isalpha(*s))
						last_token = *s - 'A' + 10;
				if (last_token == 0)
					last_token = MAXTOK;

				if (last_token < tok) {
					*(++s) = 0;
					fprintf(stderr,"Invalid Specification: %s\n",start);
					exit(4);
				}
			}
			
			while(token[tok] && tok <= last_token) {
				print_token(token[tok],o,flen,w,rightjustify,otfi);
				++tok;
			}
			break;

		case 'C':	// print out character field
			expanding = 0;
			s = getfieldoffset(s,&o);
			print_token(l,o,len,w,rightjustify,otfi);
			break;

		case 'D': // Trimmed field, remove leading, trailing whitespace
			expanding = 0;
			s=getfieldoffset(s,&o);			
			if (w == 0) {
				// If no length specified, put out entire field			
				w = len;
				if (len < (w+o))
					w = len - o;
				if (w > 0) {
					strncpy(tokbuf,l+o,w);
					tokbuf[w]=0;
					fputs(trim(tokbuf),otfi);
				}
			}
			else {
				// If the field length is >= w,just write the field data
				if ((len - o) >= w) {
					strncpy(tokbuf,l+o,w);
					tokbuf[w] = 0;
					fputs(trim(tokbuf),otfi);
				}
				else {
					// write out as much data as is present
					if (o < len) {
						strncpy(tokbuf,l+o,len-o);
						tokbuf[len-o] = 0;
						fputs(trim(tokbuf),otfi);
					}
				}
			}		
#ifdef _DEBUG
			fflush(otfi);
#endif
			break;

		case 'H':
			// output the hold buffer, suitibly modified as in C or D or F
			expanding = 0;
			s = getfieldoffset(s,&o);
			print_token(holdbuf,o,ln,w,rightjustify,otfi);
			break;

		case '#':
			// output the line number
			if (w > 5 || w < 1) w = 5;
			sprintf(temp,"%c%dd",'%',w);
			fprintf(otfi,temp,lineno);
			++s;
			expanding = 0;
			break;
		default: // output the current character
			putc(*s,otfi);
			expanding = 0;
			++s;
			break;
		}
	}
	if (usenewline) 
		putc('\n',otfi);
	/* bwr(otfi,"\n",2); */
}



void Usage() {
	FILE *hlp;
#ifdef WIN32
	char tmp[_MAX_PATH],*p,cmd[_MAX_PATH];

	if ((p=getenv("tmp")) == NULL)
		if ((p = getenv("temp")) == NULL)
			GetWindowsDirectory(tmp,_MAX_PATH);
		else
			strcpy(tmp,p);
	else
		strcpy(tmp,p);
	if (tmp[strlen(tmp)-1] != '\\')
		strcat(tmp,"\\");

	strcat(tmp,"lfmt32.hlp");
	hlp = fopen(tmp,"w");

#else
	hlp = stderr;
#endif
	fprintf(hlp,"LFMT32 [%s]\n",__DATE__);
	fprintf(hlp,"(c) Copyright Norm London, London Computing, 1987,1988,1992,1994,1997-2004\n");
	fprintf(hlp,"norm_london@londoncomputing.com\n\n");
    	fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
	fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
	fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
	fprintf(hlp,"(at your option) any later version.\n\n");
	fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
	fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	fprintf(hlp,"GNU General Public License for more details.\n\n");
	fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
	fprintf(hlp,"along with this program; if not, write to the Free Software\n");
	fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");
	fprintf(hlp,"Usage is:\nLFMT32 [-Q] [-Sn] [-D=\"xxx\"] [{-C | inputfile | NUL}] [/format] [-ooutfile] [-n] [-$]\n");
	fprintf(hlp,"For WIN32. Prints a file according to an output format.   Where:\n");
	fprintf(hlp,"inputfile -- is the name of the file to be formatted\n\n");
	fprintf(hlp,"NUL       -- specifies a file with only one empty record (CR-LF only).\n");
	fprintf(hlp,"             Allows you to use LFMT32 to output strings containing special\n");
	fprintf(hlp,"             characters, e.g. < > , | ..\n");
	fprintf(hlp,"-C        -- indicates that the file is to be read from console(STDIN).\n");
	fprintf(hlp,"             -C may be omitted if either a /format or a flag is set\n");
	fprintf(hlp,"             on the command line.\n\n");
	fprintf(hlp,"If neither an inputfile nor -C is specified, read from console(STDIN)\n");	
	fprintf(hlp,"Redirection is permitted.\n\n");
	fprintf(hlp,"-Sn         -- Skip the first n lines of the file.\n\n");
	fprintf(hlp,"-D=\"xxx\"  -- Specify a set of delimiters to separate arguments.  These\n");
	fprintf(hlp,"             are used to parse the file when a $Fn is used in the format.\n");
	fprintf(hlp,"             $F1 stands for field 1, ...  Leading and trailing Whitespace \n");
	fprintf(hlp,"             (tab, space) characters are automatically removed.  When\n");
	fprintf(hlp,"             specified any character defined between the quotes will \n");
	fprintf(hlp,"             end a field.  If omitted, a blank or tab is will end a field.\n");
	fprintf(hlp,"             Use \\> for >   \\< for <  \\| for | \\t for tab \n");
	fprintf(hlp,"                  \\n for lf \\r for cr \\\\\\\\ for \\  \\q for \"\n");
	fprintf(hlp,"             A space character is treated differently from all other\n");
	fprintf(hlp,"             delimiter characters.  Multiple spaces are treated as a single.\n");
	fprintf(hlp,"             delimiter and multiple spaces preceding some other delimiter\n");
	fprintf(hlp,"             are also treated as a single delimiter.  When the delimiter\n");
	fprintf(hlp,"             characters is not a space, multiple delimiters are treated\n");
	fprintf(hlp,"             as delimiting separate fields.  For example, if the delimiter\n");
	fprintf(hlp,"             strings is set to \" ,\" (space and comma), then the string\n");
	fprintf(hlp,"             A  B ,  C will return three fields: F1=A, F2=B and F3=C.\n");
	fprintf(hlp,"             If the delimiter string is set to \",\" (Comma Only), then the \n");
	fprintf(hlp,"             string A,,B,,C will return 5 fields:\n");
	fprintf(hlp,"             F1=A, F2=(null), F3=B, F4=(null) and F5=C\n\n");
	fprintf(hlp,"-H=\"string\"  If the input line contains the string specified, then the entire\n");
	fprintf(hlp,"             line is copied into the hold buffer.  Case is ignored.  The quotes\n");
	fprintf(hlp,"             are only necessary if the string contains whitespace characters or\n");
	fprintf(hlp,"             special characters, e.g. <,>. Within the string,\n");
	fprintf(hlp,"             Use \\> for >   \\< for <  \\| for | \\t for tab \n");
	fprintf(hlp,"                  \\n for lf \\r for cr \\\\\\\\ for \\  \\q for \"\n");
	fprintf(hlp,"-Q        -- Quiet Mode.  Do not print copyright line.\n\n");
	fprintf(hlp,"-N        -- if specified, omit newline after processing each record.\n");
	fprintf(hlp,"-%c       -- if specified, use %c instead of $.\n",'%','%');
	fprintf(hlp,"             If you select this option, you will have to double percent\n");
	fprintf(hlp,"             signs, especially useful within .BAT or .CMD files.\n\n");
	fprintf(hlp,"-ooutfile -- specifies the name of an output file.  STDOUT is default.\n\n");
#ifndef WIN32
	fprintf(hlp,"\nPlease Press RETURN\n");
	getchar();
#endif
	fprintf(hlp,"format   --  is the format specification for output\n");
	fprintf(hlp,"             Default format is: $5#: $C\n\n");
	fprintf(hlp,"Format must start with a /.  Format Characters are:\n\n");
	fprintf(hlp,"\t$[[.]n]C[+m] --Prints n characters from inputline starting at pos m.\n");
	fprintf(hlp,"\t            n defaults to the length of the line, m defaults to pos 0.\n");
	fprintf(hlp,"\t            If n is longer than the length of the field, pad with \n");
	fprintf(hlp,"\t            spaces on the right (Left Justify the field.)\n");
	fprintf(hlp,"\t            If n is preceded by a period (.), right justify the field.\n");
	fprintf(hlp,"\t$[n]D[+m]-- Like $C except whitespace trimmed from front and back of the data.\n\n");
	fprintf(hlp,"\t$[n]H[+m]-- Prints n characters from holdbuf starting at pos m.\n");
	fprintf(hlp,"\t            n defaults to the length of the line, m defaults to pos 0.\n");
	fprintf(hlp,"\t            If n is longer than the length of the field, pad with \n");
	fprintf(hlp,"\t            spaces on the left (Right Justify the field.)\n\n");
	fprintf(hlp,"\t$[n]Fk[+m][-l]--Prints n characters Field k starting at position m.\n");
	fprintf(hlp,"\t            If -l is specified, then fields k through l are printed in turn.\n");
	fprintf(hlp,"\t            If - is specified, but no ending field number, then all fields\n");
	fprintf(hlp,"\t            starting with field k to the end of the record are printed.\n");
	fprintf(hlp,"\t            n defaults to the length of the field, m defaults to pos 0.\n");
	fprintf(hlp,"\t            If n is longer than the length of the field, pad with \n");
	fprintf(hlp,"\t            spaces on the left (Right Justify the field.)\n");
	fprintf(hlp,"\t            k=1...9,a...z to specify up to 35 seperate fields.\n");
	fprintf(hlp,"\t            k can also be a pure decimal number from 1-255, allowing for\n");
	fprintf(hlp,"\t            up to 255 seperate fields.\n");
	fprintf(hlp,"\t            Each field is delimitted by one of the delimiter characters\n");
	fprintf(hlp,"\t            specified by the -D=\"xxx\" argument.  If -D= is omitted, then\n");
	fprintf(hlp,"\t            a space, tab or end of line is used to end a field.\n");
	fprintf(hlp,"\t            If a line ends before Field K is reached, a null string\n");
	fprintf(hlp,"\t            is returned.\n\n");
	fprintf(hlp,"\t$[n]N[+m]-- Prints n characters from inputline starting at pos m.\n");
	fprintf(hlp,"\t            n defaults to the length of the line, m defaults to pos 0.\n");
	fprintf(hlp,"\t            If n is longer than the length of the field, pad with \n");
	fprintf(hlp,"\t            spaces on the left (Right Justify the field.)\n\n");
	fprintf(hlp,"\t$R          Reads the next line of the file.  $C or $Fn following\n");
	fprintf(hlp,"\t            $R now refers to the new line.\n\n");
	fprintf(hlp,"\t$[n]#    -- Prints line number, justified to n positions.\n");
	fprintf(hlp,"\t            n defaults to 5.\n\n");
	fprintf(hlp,"\t$n       -- Prints n spaces.  n must be followed by a blank.\n");
	fprintf(hlp,"\t$G       -- Prints >.\n");
	fprintf(hlp,"\t$L       -- Prints <.\n");
	fprintf(hlp,"\t$B       -- Prints |.\n");
	fprintf(hlp,"\t$N       -- Prints New Line Character.\n");
	fprintf(hlp,"\t$T       -- Prints a tab character.\n");
	fprintf(hlp,"\t$Q       -- Prints a double quote (\").\n");
	fprintf(hlp,"\t$S       -- Prints a forward slash (/).\n");
	fprintf(hlp,"\t$x       -- Prints character x.  Use $+ for + following letter 'c'.\n");
	fprintf(hlp,"\t            To represent the string C+, enter C$+.\n");
	fprintf(hlp,"\t            Use $- for - anywhere in format.\n\n",'%');
	fprintf(hlp,"All other characters prints as themselves.\n\n");
	fprintf(hlp,"LFMT32 can process lines up to $d bytes long.\n\n",MAXLINE);
	fprintf(hlp,"If -%c is specified, then use %c instead of $ as shown above.\n",'%','%');
#ifndef WIN32
	fprintf(hlp,"\nPlease Press RETURN\n");
	getchar();
#endif
	fprintf(hlp,"Example:  Print the first, third and fifth through seventh characters\n");
	fprintf(hlp,"          from each line of CONFIG.SYS, with spaces between each field.\n");
	fprintf(hlp,"          LFMT32 CONFIG.SYS -$ /$1C $1C+2 $3C+4\n\n");
	fprintf(hlp,"Example:  Print the first 12 positions of the directory display of the\n");
	fprintf(hlp,"          C drive.\n");
	fprintf(hlp,"          DIR C: | LFMT32 -$ /$12C \n\n");
	fprintf(hlp,"Example:  Add a line containing <html> to stdout.\n");
	fprintf(hlp,"          LFMT32 NUL /$Lhtml$G \n\n");
	fprintf(hlp,"Example:  Print only the first 3 positions of field 2 followed by field 15\n");
	fprintf(hlp,"          from file xxx. Use a comma as a field delitter.\n");
	fprintf(hlp,"          LFMT32 -D=\",\" xxx /$3F2$Ff\n\n");
	fprintf(hlp,"Example:  Print lines 2,4,6,... of the file xxx\n");
	fprintf(hlp,"          LFMT32 xxx /$R$C\n");
#ifdef WIN32
	fclose(hlp);
	strcpy(cmd,"NOTEPAD.EXE ");
	strcat(cmd,tmp);
	if (WinExec(cmd,SW_SHOW) <= 31) {
		// An error occurred
		strcpy(cmd,"MORE <");
		strcat(cmd,tmp);
		system(cmd);
	}
#endif
}

char *stristr(char *buffer,char *pattern) {
	// returns a pointer to a location in buffer containing the pattern
	int lb = strlen(buffer);
	int lp = strlen(pattern);
	int i;
	for (i=0;i <= (lb-lp);++i) 
		if (strnicmp(buffer+i,pattern,lp)== 0)
			return buffer+i;
	return NULL;
}

void parseString(char *arg,char *str) {
	size_t j,k=2,l;
	char temp[MAXSTRPARM], inquote=0;

	// Assume no quote
	if (arg[k] == '=') // skip =, there for readability
		++k;
	if (arg[k] == '\"') { 
		// If the string starts with a quote
		inquote = 1;
		++k;
	}
	if (k < strlen(arg))
		// If a string has been specified, leave empty
		strcpy(str,arg + k);
	// remove ending quote if necessary
	if (inquote && str[strlen(str)-1] == '\"')
		str[strlen(str)-1] = 0;
		// Allow escape characters to be used
		// \t for tab, \n for lf, \r for cr, \x for x

	if (strchr(str,'\\')) {
		// look for excape characters
		strcpy(temp,str);
		*str = 0;
		// look for escaped characters in the str
		for (j=0,l=0; j < strlen(temp); ++ j) {
			if (temp[j] == '\\') {
				++j;
				switch (toupper(temp[j])) {
				case 'Q':
					temp[j] = '\"';
					break;
				case 'T':
					temp[j] = 9;
					break;
				case 'N':
					temp[j] = 10;
					break;
				case 'R':
					temp[j] = 13;
					break;
				}
			}
			str[l++] = temp[j];
		}
		str[l] = 0;
	}
	// fprintf(stderr,"Delim=\"%s\"\n",delim);
}


void main(argc,argv)
int argc;
char *argv[];
{
	static char defaultfmt_c[] = "$#: $C";
	static char defaultfmt_d[] = "$#: $F=";
	char *defaultfmt=defaultfmt_c;
	int i,skip=0;
	char infmt = 0, quiet = 0, inname[_MAX_PATH];
	char delim[MAXSTRPARM];
	holdbuf[0] = 0; //initially empty
	holdpattern[0] = 0; //initiall empty
	strcpy(delim," \t\n");
	otfi = stdout;
	strcpy(format,defaultfmt);
	strcpy(inname,"STDIN");

	if (argc < 2) {
		Usage();
		exit(0);
	}
	_fmode = _O_BINARY;

	infi = NULL;
	for (i=1;i < argc;++i) {
		switch (argv[i][0]) {
		case '/':
			infmt = 1;
			strcpy(format,argv[i]+1);
			break;
		case '-':
			switch (toupper(argv[i][1])) {
			// Skip one or more lines
			case 'S':
				skip = atoi(argv[i]+2);
				break;
			// Hold pattern.  Whenever this pattern is found the 
			// record containing the pattern is saved in the hold buffer
			case 'H':
				parseString(argv[i],holdpattern);
				break;
			// specify delimiter
			case 'D':
				defaultfmt=defaultfmt_d;
				strcpy(format,defaultfmt);
				parseString(argv[i],delim);
				break;
			// Quiet mode
			case 'Q':
				quiet = 1;
				break;
			// Change command character to %
			case '%':
				cmdchar = '%';
				if (strcmp(format,defaultfmt) == 0) {
					format[0] = '%';
					format[4] = '%';
				}
				defaultfmt[0] = '%';
				defaultfmt[4] = '%';
				break;
			// Change command character to $
			case '$':
				cmdchar = '$';
				if (strcmp(format,defaultfmt) == 0) {
					format[0] = '$';
					format[4] = '$';
				}
				defaultfmt[0] = '$';
				defaultfmt[4] = '$';
				break;
			// Do not write lf at the end of each line
			// depend on the format to end lines
			case 'N':
				usenewline = 0;
				break;
			// output file name
			case 'O':
				otfi = fopen(argv[i]+2,"w");
				if (otfi != NULL)
					fprintf(stderr,"Writing: %s\n",argv[i]+2);
				else {
					fprintf(stderr,"Unable to open output file: %s\nWriting to standard output.\n",argv[i]+2);
					otfi = stdout;
				}
				break;
			// force input from console
			case 'C':
				infi = stdin;
				infmt = 2; /* force all other input to be ignored until / */
				break;
			// infmt has three values: 0 initially
			// 1 -- accumulating format
			// 2 --
			default:
				if (infmt == 1) {
					strcat(format," ");
					strcat(format,argv[i]);
				}
				else 
					fprintf(stderr,"The following text has been ignored:\n%s\n",argv[i]);

			}
			break;
		default:
			switch (infmt) {
			case 0: /* name of input file */
				strcpy(inname,argv[i]);
				if (stricmp(argv[i],"NUL")) {
					infi = fopen(argv[i],"r");
					if (infi == NULL) {
						fprintf(stderr,"Unable to open: %s\n",argv[i]);
						exit(4);
					}
				}
				else 
					infi = NULL;
				infmt = 2;
				break;
			case 1: /* add to format line */
				strcat(format," ");
				strcat(format,argv[i]);
				break;
			default:
				fprintf(stderr,"The following text has been ignored:\n%s\n",argv[i]);
			}						
			break;
		}
	}
	if (!quiet)
		fprintf(stderr,
		"LFMT32(c) Copyright London Computing, 1987,1988,1992,1994,1997-1999 [%s]\n",__DATE__);



	if (infi == NULL) {
		if (stricmp(inname,"NUL")==0) {
			// Output a single line
			line[0]= 0;
			lineno = 1;
			expand(format,line,delim);
			if (otfi != NULL) 
				fclose(otfi);
			exit(0);
		}
		infi = stdin;
	}
	
	_setmode(_fileno(infi),_O_BINARY);
//	_setmode(_fileno(otfi),_O_BINARY);
	setvbuf(otfi,outbuff,0,sizeof(outbuff));
	setvbuf(infi,inbuff,0,sizeof(inbuff));
	
	while (skip-- && !done)
		if (fgets(line,sizeof(line),infi) == NULL)
			done = 1;

	while (!done && fgets(line,sizeof(line),infi) != NULL) {
		++lineno;
		if (*holdpattern)
			if (stristr(line,holdpattern)!=NULL) {
				strcpy(holdbuf,line);
				// Remove lf character
				i = strlen(holdbuf);
				while(--i >= 0 && (holdbuf[i] == '\n' || holdbuf[i] == '\r'))
					holdbuf[i] = 0;
			}

		expand(format,line,delim);
	}

	if (infi != NULL) fclose(infi);
	if (otfi != NULL) fclose(otfi);
	exit(0);
}
